##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from unittest import TestCase

import zope.component.eventtesting
from zope.publisher.http import HTTPRequest

from zope.errorview.http import ExceptionViewBase
from zope.errorview.interfaces import IHandleExceptionEvent


class TestEvent(TestCase):
    def setUp(self):
        zope.component.eventtesting.setUp()
        self.request = HTTPRequest('', {})

    def test_event(self):
        ExceptionViewBase(Exception(), self.request)()
        event = zope.component.eventtesting.getEvents()[0]
        self.assertEqual(event.request, self.request)
        self.assertTrue(IHandleExceptionEvent.providedBy(event))



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from unittest import TestCase

from zope.authentication.interfaces import IAuthentication
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.testlayer import ZCMLFileLayer
from zope.interface import implementer
from zope.publisher.browser import TestRequest
from zope.publisher.defaultview import getDefaultViewName
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces.http import IHTTPException
from zope.security.interfaces import Unauthorized

import zope.errorview


browser_layer = ZCMLFileLayer(zope.errorview.tests)


class MockPrincipal:
    id = 'mock principal'


@implementer(IAuthentication)  # this is a lie.
class MockAuthenticationUtility:

    status = None

    def unauthorized(self, principal_id, request):
        if self.status is not None:
            request.response.setStatus(self.status)


class TestErrorViews(TestCase):

    layer = browser_layer

    def setUp(self):
        self.request = TestRequest()

    def test_defaultname(self):
        self.assertEqual(
            getDefaultViewName(Exception(), self.request), 'index.html')
        error = NotFound(object(), self.request)
        self.assertEqual(
            getDefaultViewName(error, self.request), 'index.html')
        self.assertEqual(
            getDefaultViewName(
                Unauthorized(), self.request), 'index.html')

    def test_exceptionview(self):
        view = getMultiAdapter((Exception(), self.request), name='index.html')
        self.assertEqual(view(), 'A system error occurred.')
        self.assertEqual(self.request.response.getStatus(), 500)

    def test_notfoundview(self):
        error = NotFound(object(), self.request)
        view = getMultiAdapter((error, self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), 'The requested resource can not be found.')
        self.assertEqual(self.request.response.getStatus(), 404)

    def test_unauthorizedview(self):
        self.request.setPrincipal(MockPrincipal())
        view = getMultiAdapter(
            (Unauthorized(), self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(
            view(), 'Access to the requested resource is forbidden.')
        self.assertEqual(self.request.response.getStatus(), 403)

        getUtility(IAuthentication).status = 401
        self.assertEqual(
            view(), 'Access to the requested resource is forbidden.')
        self.assertEqual(self.request.response.getStatus(), 401)

        getUtility(IAuthentication).status = 302
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 302)

        getUtility(IAuthentication).status = 303
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 303)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from zope.authentication.interfaces import IAuthentication
from zope.component import getUtility
from zope.i18n import translate
from zope.publisher.browser import BrowserPage

from zope.errorview.http import ExceptionViewBase
from zope.errorview.http import UnauthorizedView as HTTPUnauthorizedView
from zope.errorview.i18n import _


class ExceptionView(ExceptionViewBase, BrowserPage):

    def render(self):
        msg = _('A system error occurred.')
        return translate(msg, context=self.request, default=msg)


class NotFoundView(ExceptionViewBase, BrowserPage):

    def update(self):
        self.request.response.setStatus(404)

    def render(self):
        msg = _('The requested resource can not be found.')
        return translate(msg, context=self.request, default=msg)


class UnauthorizedView(HTTPUnauthorizedView, BrowserPage):

    def update(self):
        # Set the error status to 403 (Forbidden) in the case when we
        # don't challenge the user.
        self.request.response.setStatus(403)
        # Make sure that the response is not cacheable.
        self.request.response.setHeader(
            'Expires', 'Jan, 1 Jan 1970 00:00:00 GMT')
        self.request.response.setHeader(
            'Cache-Control', 'no-store, no-cache, must-revalidate')
        self.request.response.setHeader(
            'Pragma', 'no-cache')
        self.request.response.setHeader(
            'Content-Type', 'text/plain;charset=utf-8')
        principal = self.request.principal
        getUtility(IAuthentication).unauthorized(principal.id, self.request)

    def render(self):
        if self.request.response.getStatus() not in (302, 303):
            msg = _('Access to the requested resource is forbidden.')
            return translate(msg, context=self.request, default=msg)
        return ''



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from unittest import TestCase

from zope.browser.interfaces import ISystemErrorView
from zope.component import getMultiAdapter
from zope.component import provideAdapter
from zope.component.testlayer import ZCMLFileLayer
from zope.interface import Interface
from zope.publisher.defaultview import getDefaultViewName
from zope.publisher.http import HTTPRequest
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces import TraversalException
from zope.publisher.interfaces.http import IHTTPException
from zope.publisher.interfaces.http import IHTTPRequest
from zope.publisher.interfaces.http import IMethodNotAllowed
from zope.publisher.interfaces.http import MethodNotAllowed
from zope.security.interfaces import Unauthorized

import zope.errorview
from zope.errorview import http


class TestErrorViews(TestCase):

    def setUp(self):
        self.request = HTTPRequest('', {})

    def test_exceptionviewbase(self):
        view = http.ExceptionViewBase(Exception(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(str(view), '')
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 500)

    def test_exceptionview(self):
        view = http.ExceptionView(Exception(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(str(view), '')
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 500)

    def test_systemerrormixin_view(self):
        class SystemErrorView(http.ExceptionViewBase,
                              http.SystemErrorViewMixin):
            pass
        view = SystemErrorView(Exception(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertTrue(ISystemErrorView.providedBy(view))
        self.assertTrue(view.isSystemError())
        self.assertEqual(str(view), '')
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 500)

    def test_traversalexceptionview(self):
        view = http.TraversalExceptionView(TraversalException(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 404)

        self.request = HTTPRequest('', {'REQUEST_METHOD': 'MKCOL'})
        view = http.TraversalExceptionView(TraversalException(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 404)

        self.request.setTraversalStack(['foo', 'bar'])
        view = http.TraversalExceptionView(TraversalException(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 409)

    def test_unauthorizedexceptionview(self):
        view = http.UnauthorizedView(Unauthorized(), self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 401)
        self.assertTrue(
            self.request.response.getHeader(
                'WWW-Authenticate', '', True).startswith('basic'))

    def test_methodnotallowedview(self):
        error = MethodNotAllowed(object(), self.request)
        view = http.MethodNotAllowedView(error, self.request)
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 405)
        self.assertEqual(self.request.response.getHeader('Allow'), '')

        class MyMethodNotAllowedView(http.MethodNotAllowedView):
            def allowed(self):
                return 'GET', 'POST', 'PUT', 'DELETE'

        MyMethodNotAllowedView(error, self.request)()
        self.assertEqual(self.request.response.getStatus(), 405)
        self.assertEqual(
            self.request.response.getHeader('Allow'), 'GET, POST, PUT, DELETE')


http_layer = ZCMLFileLayer(zope.errorview, zcml_file='http.zcml')


class TestErrorViewsFunctional(TestCase):

    layer = http_layer

    def setUp(self):
        self.request = HTTPRequest('', {})

    def test_defaultname(self):
        self.assertEqual(
            getDefaultViewName(Exception(), self.request), 'index.html')
        self.assertEqual(
            getDefaultViewName(
                TraversalException(), self.request), 'index.html')
        self.assertEqual(
            getDefaultViewName(
                Unauthorized(), self.request), 'index.html')
        error = MethodNotAllowed(object(), self.request)
        self.assertEqual(
            getDefaultViewName(error, self.request), 'index.html')

    def test_exceptionview(self):
        view = getMultiAdapter((Exception(), self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 500)

    def test_traversalexceptionview(self):
        view = getMultiAdapter(
            (TraversalException(), self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 404)
        # XXX test the MKCOL verb here too.

    def test_notfound(self):
        view = getMultiAdapter((NotFound(object(), self.request),
                                self.request),
                               name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 404)

    def test_unauthorizedexceptionview(self):
        view = getMultiAdapter(
            (Unauthorized(), self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 401)
        self.assertTrue(
            self.request.response.getHeader(
                'WWW-Authenticate', '', True).startswith('basic'))

    def test_methodnotallowedview(self):
        error = MethodNotAllowed(object(), self.request)
        view = getMultiAdapter((error, self.request), name='index.html')
        self.assertTrue(IHTTPException.providedBy(view))
        self.assertEqual(view(), '')
        self.assertEqual(self.request.response.getStatus(), 405)
        self.assertEqual(self.request.response.getHeader('Allow'), '')

        class MyMethodNotAllowedView(http.MethodNotAllowedView):
            def allowed(self):
                return 'GET', 'POST', 'PUT', 'DELETE'

        provideAdapter(
            MyMethodNotAllowedView,
            (IMethodNotAllowed, IHTTPRequest), Interface, 'index.html')

        view = getMultiAdapter((error, self.request), name='index.html')()
        self.assertEqual(self.request.response.getStatus(), 405)
        self.assertEqual(
            self.request.response.getHeader('Allow'), 'GET, POST, PUT, DELETE')



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from zope.browser.interfaces import ISystemErrorView
from zope.event import notify
from zope.interface import implementer
from zope.publisher.interfaces.http import IHTTPException

from zope.errorview.interfaces import HandleExceptionEvent


@implementer(ISystemErrorView)
class SystemErrorViewMixin:
    """An optional mixin to indicate a particular error view to be an "system
    error" view. This indicates the publication object to log the error again
    with the error reporting utility.

    """

    def isSystemError(self):
        return True


@implementer(IHTTPException)
class ExceptionViewBase:

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def update(self):
        self.request.response.setStatus(500)
        self.request.response.setHeader(
            'Content-Type', 'text/plain;charset=utf-8')

    def render(self):
        return ''

    def __call__(self):
        notify(HandleExceptionEvent(self.request))
        self.update()
        return self.render()

    def __str__(self):
        return self()


class ExceptionView(ExceptionViewBase):
    pass


class TraversalExceptionView(ExceptionViewBase):

    def update(self):
        super().update()
        if self.request.method == 'MKCOL' and self.request.getTraversalStack():
            # MKCOL with non-existing parent.
            self.request.response.setStatus(409)
        else:
            self.request.response.setStatus(404)


class UnauthorizedView(ExceptionViewBase):

    def update(self):
        super().update()
        self.request.unauthorized('basic realm="Zope"')
        self.request.response.setStatus(401)


class MethodNotAllowedView(ExceptionViewBase):

    # XXX define an interface for MethodNotAllowedView components.

    def allowed(self):
        # XXX how to determine the allowed HTTP methods?  XXX we need
        # a safe way to determine the allow HTTP methods. Or should we
        # let the application handle it?
        return []

    def update(self):
        super().update()
        allow = self.allowed()
        self.request.response.setStatus(405)
        self.request.response.setHeader('Allow', ', '.join(allow))



from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implementer


class IHandleExceptionEvent(Interface):
    """Event fired when handling an exception."""

    request = Attribute('The current request')


@implementer(IHandleExceptionEvent)
class HandleExceptionEvent:

    def __init__(self, request):
        self.request = request
